﻿using UnityEngine;
using AudioSystem;
using System.Collections;
using System.Collections.Generic;

/// <summary>
/// The Audio System manager all the sounds in the game.
/// The system upon request creates the nessecery objects to play the sounds therefore the in game elements dont need to have separate Audio Sources.
/// </summary>
namespace AudioSystem
{
	/// <summary>
	/// Audio manager is the main handler for the AudioSystem where we can define a list with all the sounds and music available in the scene
	/// </summary>
	public class AudioManager : MonoBehaviour
	{

		[Header ("General Settings")]
		/// <summary>
				/// Is the player muter
				/// </summary>
				public bool
			isMuted = false;

		[Header ("Music List")]
		/// <summary>
				/// The music list.
				/// </summary>
				public List<AudioClass>
			musicList = new List<AudioClass> ();
		[Header ("SoundFx List")]
		/// <summary>
				/// The Sound FX list.
				/// </summary>
				public List<AudioClass>
			audioList = new List<AudioClass> ();

		[HideInInspector]
		public List<GameObject>
			audioObjList = new List<GameObject> ();
		
		// Current Music Sound
		private GameObject currentMusicObj;
		

		public static AudioManager _instance;

		public static AudioManager Instance {
			get {
				return _instance;
			}
		}

		public void Awake ()
		{
			_instance = this;
		}

		
		//--------------------play Sfx
		/// <summary>
		/// Play sfx.
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="pitch">Pitch.</param>
		public void AudioPlaySfx (string _id, float delay = 0, float pitch = 1)
		{ 
			if (isMuted)
				return;

						

			StartCoroutine (PlaySound (_id, delay, pitch));
		}

		/// <summary>
		/// Plays the sound.
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="pitch">Pitch.</param>
		IEnumerator PlaySound (string _id, float delay, float pitch = 1)
		{
			yield return StartCoroutine (Auto.Wait (delay));
			foreach (AudioClass audioItem in audioList) {
				if (audioItem.id == _id) {
					AudioHelper.CreatePlayAudioObject (audioItem.audioClip, audioItem.volume, audioItem.audioClip.name, pitch);
										
					if (TextHandler.Instance != null) {
						string dspTxt = TextHandler.Instance.GetText (_id);
						Debug.Log (string.Format ("dspTxt : {0}", dspTxt));
						if (dspTxt != _id) {
							TopBarController.ShowBar (dspTxt);

							Debug.Log (string.Format ("audioItem.audioClip.length : {0}", audioItem.audioClip.length));
							StartCoroutine (RemoveTopBarTxt (audioItem.audioClip.length + 0.25f));
						}

					}

					break;
				}
			}
		}

		IEnumerator RemoveTopBarTxt (float time)
		{
			yield return new WaitForSeconds (time);
			Debug.Log (string.Format ("RemoveTopBarTxt : {0}", time));
			TopBarController.HideBar ();
		}

		/// <summary>
		/// Static Function Plays the sfx.
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="pitch">Pitch.</param>
		public static void PlaySfx (string _id, float delay = 0, float pitch = 1)
		{
			if (Instance)
				Instance.AudioPlaySfx (_id, delay, pitch);
		}



		//--------------------Play Loop Sound
		/// <summary>
		/// Play's a loop sound.
		/// </summary>
		/// <returns>The play loop sound.</returns>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="isLoop">If set to <c>true</c> is loop.</param>
		public GameObject AudioPlayLoopSound (string _id, float delay = 0, bool isLoop = false)
		{
			//Debug.Log (string.Format ("AudioPlayLoopSound: {0}", _id));

			GameObject newAudioObj = null;
		
			foreach (AudioClass audioItem in audioList) {
				if (audioItem.id == _id) {
					newAudioObj = AudioHelper.CreateGetFadeAudioObject (audioItem.audioClip, isLoop, null, _id);
					AudioSource audioObj = newAudioObj.GetComponent<AudioSource> ();
					audioObj.volume = audioItem.volume;
					audioObj.Play ((ulong)delay);
					audioObjList.Add (newAudioObj);
					break;
				}
			}
			return newAudioObj;

			//StartCoroutine (PlayLpSnd (_id, delay, isLoop));
		}

		/// <summary>
		/// Plays a loop snd.
		/// </summary>
		/// <returns>Game Object with the audio source</returns>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="isLoop">If set to <c>true</c> is loop.</param>
		IEnumerator PlayLpSnd (string _id, float delay = 0, bool isLoop = false)
		{
			yield return StartCoroutine (Auto.Wait (delay));
				
			GameObject newAudioObj;
		
			foreach (AudioClass audioItem in audioList) {
				if (audioItem.id == _id) {
					newAudioObj = AudioHelper.CreateGetFadeAudioObject (audioItem.audioClip, isLoop, null, _id);
					AudioSource audioObj = newAudioObj.GetComponent<AudioSource> ();
					audioObj.volume = audioItem.volume;
					audioObj.Play ();
					audioObjList.Add (newAudioObj);
					break;
				}
			}
		}

		/// <summary>
		/// Plays the loop sound.
		/// </summary>
		/// <returns>The loop sound.</returns>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		/// <param name="isLoop">If set to <c>true</c> is loop.</param>
		public static GameObject PlayLoopSound (string _id, float delay = 0, bool isLoop = false)
		{
			if (Instance)
				return Instance.AudioPlayLoopSound (_id, delay, isLoop);
			else
				return null;
		}

		// Stop Loop Sound
		/// <summary>
		/// Stop's a looped sound
		/// </summary>
		/// <param name="_id">Identifier.</param>
		public void AudioStopLoopSound (string _id)
		{
			//StartCoroutine (StopLpdSnd (_id));

			for (int i = 0; i < audioObjList.Count; i++) {
				if (audioObjList [i].name == _id) {
					//Debug.Log (string.Format ("StopLpdSnd {0} Found", _id));
					Destroy (audioObjList [i]);
					audioObjList.RemoveAt (i);				
					break;
				}
				
			}

		}

		/*
		IEnumerator  StopLpdSnd (string _id)
		{
				yield return StartCoroutine (Auto.Wait (0.2f));

				for (int i = 0; i < audioObjList.Count; i++) {
						if (audioObjList [i].name == _id) {
								Debug.Log (string.Format ("StopLpdSnd {0} Found", _id));
								Destroy (audioObjList [i]);
								audioObjList.RemoveAt (i);				
								break;
						}
				}
		}
		*/
		 
		
		/// <summary>
		/// Stops the looped sound.
		/// </summary>
		/// <param name="_id">Identifier.</param>
		public static void StopLoopedSound (string _id)
		{
			//Debug.Log (string.Format ("StopLoopedSound: {0}", _id));
			if (Instance)
				Instance.AudioStopLoopSound (_id);
		}

		/// <summary>
		/// Changes Loop sound Volume
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="vol">Vol.</param>
		public void LpdSetVol (string _id, float vol)
		{
			for (int i = 0; i < audioObjList.Count; i++) {
				if (audioObjList [i].name == _id) {
					AudioSource audioObj = audioObjList [i].GetComponent<AudioSource> ();			
					audioObj.volume = vol;
					break;
				}
			}
		}

		/// <summary>
		/// Set the vilume on a loop sound
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="vol">Vol.</param>
		public static void LoopedSetVol (string _id, float vol)
		{
			if (Instance)
				Instance.LpdSetVol (_id, vol);
		}
		

	
		//--------------------Music
		/// <summary>
		/// Play an audio music file
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		public void AudioPlayMusic (string _id, float delay = 0)
		{
			//Debug.Log (string.Format ("AudioPlayMusic: {0}", _id));
			if (isMuted)
				return;
			StartCoroutine (PlayMsc (_id, delay));
		}

		/// <summary>
		/// Plays a music file.
		/// </summary>
		/// <returns>The msc.</returns>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		IEnumerator PlayMsc (string _id, float delay)
		{
			yield return StartCoroutine (Auto.Wait (delay));
			///-----------------------------------------------Insert Music Check here
			foreach (AudioClass musicItem in musicList) {
				if (musicItem.id == _id) {
					currentMusicObj = AudioHelper.CreateGetFadeAudioObject (musicItem.audioClip, true);
					AudioSource musicObj = currentMusicObj.GetComponent<AudioSource> ();
					musicObj.volume = musicItem.volume;
					musicObj.Play ();
					//Debug.Log (string.Format ("currentMusicObj {0}", currentMusicObj));
					break;
				}
			}
		}

		/// <summary>
		/// Plays a music file.
		/// </summary>
		/// <param name="_id">Identifier.</param>
		/// <param name="delay">Delay.</param>
		public static void PlayMusic (string _id, float delay = 0)
		{
			//Debug.Log (string.Format ("PlayMusic: {0}", _id));
			if (Instance)
				Instance.AudioPlayMusic (_id, delay);
		}


	}
}