﻿using UnityEngine;
using AudioSystem;
using UnityEngine.Analytics;
 
using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

/// <summary>
/// Room controler is the root for every scene.
/// It handles every step of the games level timeline, the level characters, scores and core modules
/// </summary>
public class RoomControler : MonoBehaviour
{

	[Header ("Scenario Settings")]
	/// <summary>
		/// Currents scene scenario data holder
		/// </summary>
		public RoomScenario
		scenario;

	/// <summary>
	/// Ideal time for the player to finish the game.
	/// Every second above result in extra score
	/// </summary>
	public float bounusTime;
	public float bonusTimeMultiplier = 10;

	/// <summary>
	/// The scenario list (Level timeline)
	/// </summary>
	private List<RoomAction> scenarioList;

	/// <summary>
	/// How much time for the class to wait before starting the time line
	/// </summary>
	public float startDelay = 0.5f;
	/// <summary>
	/// The current scenario step;
	/// </summary>
	public int currentScenarioID = 0;
	/// <summary>
	/// The current scenario Action
	/// </summary>
	private RoomAction currentScenario;
	/// <summary>
	/// The current game object that in action
	/// </summary>
	public GameObject currentGameObjectAction;

	/// <summary>
	/// List of all the character currently active in the scene
	/// </summary>
	//List<CharacterClass> characterList;

	[Header ("Main Utilities")]

	/// <summary>
		/// The top bar module that display usefull information for the game
		/// </summary>
		public TopBarController topBar;


	/// <summary>
	/// A tool tha handles object transition like a slide up or a zoom out.
	/// </summary>
	public TransitionHandler transHandler;

	/// <summary>
	/// A tool tha handles cameras disolves to black
	/// </summary>
	public CameraFade camFader;



	[HideInInspector]
	public bool
		isMale = true;

	public static bool _isMale = true;
	

	private static RoomControler
		_instance;

	public static RoomControler Instace {
		get {
			return _instance;
		}
	}


	

	#region  Initial Functions

	public void Awake ()
	{
		_instance = this;
	
	}

	public void Start ()
	{
			

		/// Get the main scenario list (Timeline) and character of the scene
		scenarioList = scenario.scenarioList;
		//characterList = scenario.characterList;


		StartLevel ();
	}

	/// <summary>
	/// Starts the level timeline with desired delay
	/// </summary>
	public void StartLevel ()
	{
		StartCoroutine (StartScenarion (startDelay));
	}

	#endregion

	                                   
	#region Scenario Interation

	private bool musicIsPlaying = false;

	/// <summary>
	/// Starts the level main timeline
	/// </summary>
	/// <param name="delayTime">initial Delay time.</param>
	IEnumerator StartScenarion (float delayTime)
	{
		yield return StartCoroutine (Auto.Wait (delayTime));

		//Start the first Scenario Step
		NextSequence ();

		//if no music is playing start main theme
		if (!musicIsPlaying) {
			AudioManager.PlayMusic ("Theme1");
			musicIsPlaying = true;
		}
	}

	/// <summary>
	/// Handles the next step in the level Timeline
	/// </summary>
	void NextSequence ()
	{
		//	if the timeline is no completed go to next step
		if (currentScenarioID < scenarioList.Count) {
			StartCoroutine (HandleNextSection (scenarioList [currentScenarioID]));
			currentScenarioID++;
		} else {
			// else complete the level
			RoomCompleted ();
		}
	}

	/// <summary>
	/// First part of Timeline handler.
	/// </summary>
	/// <param name="_currentScenario">Current scenario item</param>
	IEnumerator HandleNextSection (RoomAction _currentScenario)
	{
		// check if the room action (timeline step) is Active and not an group Container
		if (_currentScenario.isActive && _currentScenario.taskV2 != RoomTasksV2.GroupStart && _currentScenario.taskV2 != RoomTasksV2.GroupEnd) {
			currentScenario = _currentScenario;
			// move to proccess the next step with desired waiting time.
			StartCoroutine (NextSectionDelayHandler (currentScenario.waitTime));
		} else {
			//if the room action is part of a group of an inactive group interacte throw the timeline until we find the end of the group
			if (_currentScenario.taskV2 == RoomTasksV2.GroupStart && !_currentScenario.isActive) {
				for (int i = currentScenarioID; i < scenarioList.Count; i++) {
					if (scenarioList [i].taskV2 == RoomTasksV2.GroupEnd) {
						currentScenarioID = i - 1;
						break;
					}
				}
			}		
			yield return new WaitForEndOfFrame ();
			NextSequence ();
		}
		
	}

	#endregion


	#region NextSectionDelayHandler

	/// <summary>
	/// Second part of timeline handler.
	/// </summary>
	/// <returns>The section delay handler.</returns>
	/// <param name="delayTime">Delay time.</param>
	IEnumerator NextSectionDelayHandler (float delayTime)
	{
		// wait for next step as define in the timeline
		yield return StartCoroutine (Auto.Wait (delayTime));


		// Switch case for each Timeline Step
		switch (currentScenario.taskV2) {
		case RoomTasksV2.EnableTask:
			EnableTask (currentScenario.gameItem);
			break;

		case RoomTasksV2.ShowTopBar:
			StartCoroutine (ShowTopBar (currentScenario.Text));
			break;
		case RoomTasksV2.HideTopBar:
			StartCoroutine (HideTopBar ());
			break;
	
		case RoomTasksV2.ActivateObject:
			StartCoroutine (ActivateDeActivateObjectState (currentScenario.gameItem));
			break;
		case RoomTasksV2.DeactivateObject:
			StartCoroutine (ActivateDeActivateObjectState (currentScenario.gameItem));
			break;
		case RoomTasksV2.Transition:
			TransitionObject (currentScenario.gameItem);
			break;
		case RoomTasksV2.PlaySoundFX:
			StartCoroutine (PlaySoundFX ());
			break;

		case RoomTasksV2.CameraFadeIn:
			StartCoroutine (CameraFadeIn ());
			break;
		case RoomTasksV2.CameraFadeOut:
			StartCoroutine (CameraFadeOut ());
			break;
		}
	}

	#endregion


	#region Scenario Actions

	/// <summary>
	/// Most simple and yet most important scenario step. 
	/// Sends a simple "enable task" message to a selected game object and waits for it to send a "task complete" in order to go to the next timeline step
	/// </summary>
	/// <param name="gameItem">Game item.</param>
	private void EnableTask (GameObject gameItem)
	{
		currentGameObjectAction = gameItem;
		if (!gameItem.activeInHierarchy) {
			gameItem.SetActive (true);
			StartCoroutine (EnableDelay (gameItem));
		} else
			gameItem.SendMessage (HRUtilities.EnableTask);
	}

	IEnumerator EnableDelay (GameObject gameItem)
	{
		yield return 0;
		gameItem.SendMessage (HRUtilities.EnableTask);
	}

	





	/// <summary>
	/// Shows a top bar message
	/// </summary>
	/// <param name="Text">Text.</param>
	IEnumerator ShowTopBar (string Text)
	{
		currentGameObjectAction = topBar.gameObject;
		topBar.ShowBarMsg (Text);
		yield return  StartCoroutine (Auto.Wait (HRUtilities.at));
		TaskCompleted (topBar.gameObject);
	}

	/// <summary>
	/// Hides the top bar.
	/// </summary>
	/// <returns>The top bar.</returns>
	IEnumerator HideTopBar ()
	{
		currentGameObjectAction = topBar.gameObject;
		topBar.HideBarMsg ();
		yield return  StartCoroutine (Auto.Wait (HRUtilities.at));
		TaskCompleted (topBar.gameObject);
	}

	
	/// <summary>
	/// Activates or deactivates an object.
	/// If their more the one in the timeline list it will interate trough all of them
	/// </summary>
	/// <param name="gameItem">Game item.</param>
	IEnumerator ActivateDeActivateObjectState (GameObject gameItem)
	{
		currentGameObjectAction = gameItem;
		
		int currentid = currentScenarioID;
		int completedInt = 0;	
		
		for (int i = currentid - 1; i < scenarioList.Count; i++) {
			if (scenarioList [i].taskV2 == RoomTasksV2.DeactivateObject) {
				currentScenario = scenarioList [i];
				currentScenario.gameItem.SetActive (false);
			} else if (scenarioList [i].taskV2 == RoomTasksV2.ActivateObject) {
				currentScenario = scenarioList [i];
				currentScenario.gameItem.SetActive (true);
			} else if (scenarioList [i].taskV2 == RoomTasksV2.GroupStart || scenarioList [i].taskV2 == RoomTasksV2.GroupEnd) {
				if (!scenarioList [i].isActive) {
					completedInt = i;
					break;
				} else
					continue;
			} else {
				completedInt = i;
				break;
			}
		}
		
		currentScenarioID = completedInt;
		
		TaskCompleted (gameItem);
		yield return null; 
	}

	/// <summary>
	/// Transitions an object.
	/// </summary>
	/// <param name="gameItem">Game item.</param>
	void TransitionObject (GameObject gameItem)
	{
		currentGameObjectAction = transHandler.gameObject;
		transHandler.AnimateObject (gameItem, currentScenario.animType);
	}




	/// <summary>
	/// Plays a sound Fx.
	/// </summary>
	IEnumerator PlaySoundFX ()
	{
		currentGameObjectAction = AudioManager.Instance.gameObject;
		
		AudioManager.PlaySfx (currentScenario.Text, currentScenario.walkSpeed);


		yield return  StartCoroutine (Auto.Wait (HRUtilities.at / 2));
		
		TaskCompleted (AudioManager.Instance.gameObject);
	}


	/// <summary>
	/// Fade in the camera
	/// </summary>
	IEnumerator CameraFadeIn ()
	{
		currentGameObjectAction = camFader.gameObject;
		
		camFader.StartFade (new Color (0, 0, 0, 0), currentScenario.walkSpeed);
		yield return  StartCoroutine (Auto.Wait (currentScenario.walkSpeed * 0.5f));
		TaskCompleted (camFader.gameObject);	
	}

	/// <summary>
	/// Fade out the camera
	/// </summary>
	/// <returns>The fade out.</returns>
	IEnumerator CameraFadeOut ()
	{
		currentGameObjectAction = camFader.gameObject;
		camFader.StartFade (new Color (0, 0, 0, 1), currentScenario.walkSpeed);
		yield return  StartCoroutine (Auto.Wait (currentScenario.walkSpeed * 0.5f));
		TaskCompleted (camFader.gameObject);	
	}

	#endregion

	


	#region Completed

	/// <summary>
	/// Called when a Task is completed. (When a timeline step is complete)
	/// </summary>
	/// <param name="gameItem">Game item.</param>
	public void TaskCompleted (GameObject gameItem)
	{

		//Check to see if the complete task originates from the same gameobject other wise is incorrect
		if (gameItem == currentGameObjectAction) {
			//if the task was montitored for score set as done

			//go to next timeline step
			NextSequence ();
		} else {
			Debug.Log (string.Format ("{2}: TaskCompleted () NOT SAME Task Was Completed! gameItem:{0} - currentGameObjectAction:{1}", gameItem.name, currentGameObjectAction.name, Time.time));
		}
	}

	
	/// <summary>
	/// The Room is completed. 
	/// </summary>
	void RoomCompleted ()
	{
		Debug.Log ("RoomCompleted()");
		// Collect the time it took to complete

	}


	

	#endregion


}
