using UnityEngine;
using UnityEngine.Events;
using UnityEngine.EventSystems;

using AudioSystem;
using System.Collections;


/// <summary>
/// Dragable item is a generic class that assist in creating puzzles (tasks) with dragable objects.
/// The class works as a parent to the following childs with specific namining convantion:
///	"normal": a gameobject that will contain the normal state graphics (requierd)
///	"glow": a gameobject that will contain the glow state graphics
///	"final": a gameobject that will is the correct position of the game object
/// </summary>
public class DragableItem : MonoBehaviour , IPointerDownHandler , IPointerUpHandler
{

	public bool isActive;

	//Dragable Properties
	[Header ("Dragable Properties")]

	public bool SelfCheck = true;
	public bool contineusCheck = false;

	public bool swapAtWithFinal = true;

	public bool finalIsGuide = false;

	public bool isInCorrectPosition = false;

	public bool allowReturn = false;


	[Range (0, 10)]
	public float maxWinDistance = 5f;

	public Camera rayCam;

	public bool resizeCollider = true;
	public bool handlePositioning = true;
	//Main Objects
	[Header ("Objects Properties")]
	public GameObject SpriteObj;

	public GameObject GlowObj;

	public GameObject FinalObj;
		
	[Space (5)]
	/// <summary>
		/// if this is set when final object will become a child to this object
		/// </summary>
		public Transform
		finalParent;

	//Drag And Drop
	private BoxCollider myBoxCollider;
	private Vector3 StartColliderSize;
	private Vector3 StartPosition;
	private Vector3 mouseOffSet;
	public bool isDragged = false;

	[HideInInspector]
	public Vector3
		StartDragPosition;

	[Header ("Feedback Properties")]
	public string
		hoverText = "";
	public string goodfeedBackText = "";
	public string badfeedBackText = "";
		
	public float feedBackTime = 3f;
	public bool isHoverTextEnabled = true;


	void Awake ()
	{
		StartPosition = myTransform.position;

		//Enable the Hover events in the ui


		//Find the normal state
		if (myTransform.FindChild ("normal")) {
			SpriteObj = myTransform.FindChild ("normal").gameObject;
			SpriteObj.SetActive (true);
		}

		//Find the Glow
		if (myTransform.FindChild ("glow")) {
			GlowObj = myTransform.FindChild ("glow").gameObject;
			GlowObj.SetActive (false);
		}

		//Find The finalpos
		if (SelfCheck && myTransform.FindChild ("final")) {
			FinalObj = myTransform.FindChild ("final").gameObject;

			//Configure the parent
			if (finalParent)
				FinalObj.transform.parent = finalParent;
			else
				FinalObj.transform.parent = myTransform.parent;

			//Set inactive if its just a guide or for final placement
			if (swapAtWithFinal || finalIsGuide)
				FinalObj.SetActive (false);						
		}

		// Handle collider size etc
		if (resizeCollider) {
			myBoxCollider = GetComponent <BoxCollider> ();
			StartColliderSize = myBoxCollider.size;
			myBoxCollider.size = Vector3.zero;
		}

	}

	// Activate The Object
	public void EnableObjective ()
	{	
		// Set the collider to correct size
		if (resizeCollider)
			myBoxCollider.size = StartColliderSize;

		//Get all the text's
		if (TextHandler.Instance) {
			hoverText = TextHandler.Instance.GetText (hoverText);
			goodfeedBackText = TextHandler.Instance.GetText (goodfeedBackText);
			badfeedBackText = TextHandler.Instance.GetText (badfeedBackText);
		}

		//Fix the events
		FixEvents ();
	}



	/// <summary>
	/// On the mouse down.
	/// </summary>
	/// <param name="uitem">Uitem.</param>
	public virtual void OnPointerDown (PointerEventData eventData)
	{
		if (!isActive)
			return;
		//Capture the difference from mouse and object position
		mouseOffSet = myTransform.position - HRUtilities.GetMousePosition (rayCam);

		//sets the drag starting point
		StartDragPosition = myTransform.position;

		//Set the status in Dragable mode
		isDragged = true;


		SetColliderSizeToDrag ();
	}

	public virtual void OnPointerUp (PointerEventData eventData)
	{
		if (!isActive)
			return;
		//Set the status in normal mode
		isDragged = false;
		//Deactivate the hover
		if (GlowObj)
			GlowObj.SetActive (false);
		//Check if in correct position
		//CheckForWictory ();
		SetColliderSizeToStartSize ();
	}

	/// <summary>
	/// Checks if object is in correct position
	/// </summary>
	public void CheckForWictory ()
	{
		// if class set not to check its position then we just send an "ObjectiveCompleted" message to the task hanlder
		if (!SelfCheck) {
			gameObject.SendMessageUpwards (HRUtilities.ObjectiveCompleted, gameObject, SendMessageOptions.DontRequireReceiver);
			return;
		}


		float winDistance;
		//Different types of distance calculation depending if we checking for objects Original Position
		//or the final
		if (isInCorrectPosition && allowReturn) //Original position
						winDistance = Vector3.Distance (StartPosition, myTransform.position);
		else/// final Position
						winDistance = Vector3.Distance (FinalObj.transform.position, myTransform.position);

		// check if withing limit
		if (winDistance < maxWinDistance) {
			//if we allow object to be moved after its has been place
			if (allowReturn) {
				// if it was in the correct position we set it false and return it to its original position
				if (isInCorrectPosition) {
					isInCorrectPosition = false;
					StartCoroutine (SetToFinalPosition (StartPosition, Quaternion.identity));
				} else {
					// else we take it to its final position
					isInCorrectPosition = true;
					StartCoroutine (SetToFinalPosition (FinalObj.transform.position, FinalObj.transform.rotation));
				}
			} else {
				isInCorrectPosition = true;
				//gameObject.SendMessageUpwards (HRUtilities.ObjectiveCompleted, gameObject, SendMessageOptions.DontRequireReceiver);
				StartCoroutine (SetToFinalPosition (FinalObj.transform.position, FinalObj.transform.rotation));
			}

			//Give user goodfeedback
			if (goodfeedBackText != "")
				TopBarController.ShowObjective (goodfeedBackText, true, feedBackTime);

			myTransform.parent.SendMessageUpwards (HRUtilities.AddScore, SendMessageOptions.DontRequireReceiver);
		} else {
			//Give User Back Feedback
			if (badfeedBackText != "")
				TopBarController.ShowObjective (badfeedBackText, false, feedBackTime);
			//Remove score from player
			myTransform.parent.SendMessageUpwards (HRUtilities.RemoveScore, SendMessageOptions.DontRequireReceiver);

			// move the object to its original position
			if (handlePositioning)
				StartCoroutine (SetToStartPosition ());
		}
	}


	public void SimpleCheckForWictory ()
	{
		//Debug.Log (string.Format ("SimpleCheckForWictory ()"));
		float winDistance = Vector3.Distance (FinalObj.transform.position, myTransform.position);

		if (winDistance < maxWinDistance) {
			isInCorrectPosition = true;
			contineusCheck = false;
			isActive = false;
			Debug.Log (string.Format ("SimpleCheckForWictory Complete : {0}", this.name));
			//StartCoroutine (SetToFinalPosition (FinalObj.transform.position, FinalObj.transform.rotation));
			gameObject.SendMessageUpwards (HRUtilities.ObjectiveCompleted, gameObject, SendMessageOptions.DontRequireReceiver);
		}
	}

	/// <summary>
	/// Moves to final position.
	/// </summary>
	/// <param name="pos">Position.</param>
	/// <param name="sendObjectiveComplete">If set to <c>true</c> send objective complete.</param>
	public void MoveToFinalPosition (Vector3 pos, bool sendObjectiveComplete = true)
	{
		StartCoroutine (SetToFinalPosition (pos, Quaternion.identity, sendObjectiveComplete));
	}

	/// <summary>
	/// Move object to its final position.
	/// </summary>
	/// <param name="pos">Position.</param>
	/// <param name="rot">Rot.</param>
	/// <param name="sendObjectiveComplete">If set to <c>true</c> send objective complete.</param>
	public IEnumerator SetToFinalPosition (Vector3 pos, Quaternion rot, bool sendObjectiveComplete = true)
	{
		Vector3 localFinalPos = pos;
		//Animate movement
		StartCoroutine (myTransform.MoveToGlobal (localFinalPos, HRUtilities.at, EaseType.QuadOut));
		StartCoroutine (myTransform.RotateToGlobal (rot, HRUtilities.at, EaseType.QuadOut));

		if (SelfCheck) {
			yield return StartCoroutine (Auto.Wait (0.25f));
			// Swap the normal and final state graphics
			if (swapAtWithFinal) {
				gameObject.SetActive (false);
				if (!finalIsGuide)
					FinalObj.SetActive (true);
			}
			// send objective complete message
			if (sendObjectiveComplete)
				gameObject.SendMessageUpwards (HRUtilities.ObjectiveCompleted, gameObject, SendMessageOptions.DontRequireReceiver);
		}
	}

	/// <summary>
	/// Moves to start position.
	/// </summary>
	public void MoveToStartPosition ()
	{
		StartCoroutine (SetToStartPosition ());
	}

	/// <summary>
	/// Move object to its starting position.
	/// </summary>
	public IEnumerator SetToStartPosition ()
	{
		if (allowReturn) {

			if (isInCorrectPosition) {
				isInCorrectPosition = false;
				yield return StartCoroutine (myTransform.MoveToGlobal (FinalObj.transform.position, HRUtilities.at, EaseType.QuadOut));
			} else {
				isInCorrectPosition = true;
				yield return StartCoroutine (myTransform.MoveToGlobal (StartPosition, HRUtilities.at, EaseType.QuadOut));
			}

		} else
			yield return StartCoroutine (myTransform.MoveToGlobal (StartPosition, HRUtilities.at, EaseType.QuadOut));

	}

	
	/// <summary>
	/// Fixs the events.
	/// </summary>
	public virtual void FixEvents ()
	{

		Debug.Log (string.Format ("FixEvents : {0}", this.name));
		//UIItem.OnDownUIItem += onMouseDown;
		isActive = true;
	
	}

	/// <summary>
	/// Removes the events.
	/// </summary>
	public virtual void RemoveEvents ()
	{
		if (!allowReturn)
			isActive = false;


		if (GlowObj)
			GlowObj.SetActive (false);
	}

	/// <summary>
	/// Move The Object when dragged mode is enabled
	/// </summary>
	public virtual void FixedUpdate ()
	{
		if (!isDragged)
			return;
		
		Vector3 MousePos = HRUtilities.GetMousePosition (rayCam);
		myTransform.position = new Vector3 (MousePos.x + mouseOffSet.x, MousePos.y + mouseOffSet.y, myTransform.position.z);
	}

	/// <summary>
	/// Resets the object state
	/// </summary>
	/// <param name="sendToOriginalPosition">If set to <c>true</c> moves the object to the  original position.</param>
	public void Reset (bool sendToOriginalPosition = true)
	{
		gameObject.SetActive (true);
		FinalObj.SetActive (false);

		if (sendToOriginalPosition) {
			MoveToStartPosition ();

			SetColliderSizeToStartSize ();
			RemoveEvents ();
			FixEvents ();
		}
	}

	/// <summary>
	/// Sets the collider size to drag.
	/// </summary>
	private void SetColliderSizeToDrag ()
	{
		if (resizeCollider)
			myBoxCollider.size = Vector3.one * HRUtilities.colliderDragScale;
		
	}

	/// <summary>
	/// Sets the size of the collider size to start.
	/// </summary>
	private void SetColliderSizeToStartSize ()
	{
		if (resizeCollider)
			myBoxCollider.size = StartColliderSize;
	}


	private Transform _myTransform;

	public  Transform myTransform {
		get {
			if (_myTransform == null)
				_myTransform = GetComponent<Transform> ();
			return _myTransform;
		}
		
		set {
			_myTransform = value;
		}
		
	}


	
}