﻿using UnityEngine;
using UnityEngine.Events;
using UnityEngine.EventSystems;
using AudioSystem;
using System.Collections;

/// <summary>
/// The Dragable physics class extends the DragableItem. 
/// It was created for the maze mini game in the production game in order to have a dragable object that reacted to Collitions
/// </summary>
public class DragablePhysics : DragableItem
{


	[Header ("Physics Settings")]
	public int normalCollisionCount = 1;
	public float moveLimit = .5f;
	public float collisionMoveFactor = .01f;
	public float addHeightWhenClicked = 0.0f;
	public bool freezeRotationOnDrag = true;
	public Camera cam;
	//private Transform myTransform  ;
	private bool canMove = false;
	private float zPos;
	private bool gravitySetting;
	//private bool freezeRotationSetting;
	private float sqrMoveLimit;
	public int collisionCount = 0;
	private Transform camTransform;

	public Vector3 offset = new Vector3 ();

	public bool applyReleaseForce = false;

	private Vector3 prevPos;
	[Range (0, 100)]
	public float
		forceMultiplyer;

	public bool reverseCamera = false;

	private Rigidbody _myRigidbody;

	public Rigidbody myRigidbody {
		get {
			if (_myRigidbody == null)
				_myRigidbody = GetComponent<Rigidbody> ();
			return _myRigidbody;
		}
		set {
			_myRigidbody = value;
		}
	}



	void Start ()
	{
		if (!cam) {
			cam = Camera.main;
		}
		if (!cam) {
			Debug.LogError ("Can't find camera tagged MainCamera");
			return;
		}
		camTransform = cam.transform;
		sqrMoveLimit = moveLimit * moveLimit;   // Since we're using sqrMagnitude, which is faster than magnitude
	}


	public override void OnPointerDown (PointerEventData eventData)
	{
		base.OnPointerDown (eventData);
	
		if (!isActive)
			return;
		Debug.Log (string.Format ("DragablePhysics - onMouseDown ({0})", isActive));

		canMove = true;
		myTransform.Translate (Vector3.up * addHeightWhenClicked);
		gravitySetting = myRigidbody.useGravity;
		//freezeRotationSetting = myRigidbody.freezeRotation;
		myRigidbody.useGravity = false;
		//myRigidbody.freezeRotation = freezeRotationOnDrag;
		zPos = myTransform.position.z;
	}

	public override void OnPointerUp (PointerEventData eventData)
	{
		base.OnPointerUp (eventData);

		Debug.Log (string.Format ("DragablePhysics - onMouseUP ({0})", isActive));
		if (!isActive)
			return;

		canMove = false;
		myRigidbody.useGravity = gravitySetting;
		//myRigidbody.freezeRotation = freezeRotationSetting;
		if (!myRigidbody.useGravity) {
			Vector3 pos = myTransform.position;
			pos.z = zPos - addHeightWhenClicked;
			myTransform.position = pos;
		}

		Vector3 diff = myRigidbody.position - prevPos;
		myRigidbody.AddForce (diff * forceMultiplyer, ForceMode.Impulse);
	}

	void OnCollisionEnter ()
	{
		collisionCount++;
	}

	void OnCollisionExit ()
	{
		collisionCount--;
	}

	public override void FixedUpdate ()
	{

		if (contineusCheck && isActive && !canMove)
			SimpleCheckForWictory ();


		if (!canMove)
			return;
		

		prevPos = myRigidbody.position;

		myRigidbody.velocity = Vector3.zero;
		myRigidbody.angularVelocity = Vector3.zero;
		
		Vector3 pos = myTransform.position;
		pos.z = zPos;
		myTransform.position = pos;
		
		Vector3 mousePos = Input.mousePosition;

		Vector3 mouseConversion;

		if (!reverseCamera)
			mouseConversion = new Vector3 (mousePos.x, mousePos.y, camTransform.position.z - myTransform.position.z);
		else
			mouseConversion = new Vector3 (mousePos.x, mousePos.y, myTransform.position.z - camTransform.position.z);

		Vector3 move = cam.ScreenToWorldPoint (mouseConversion) - myTransform.position;
		move.z = 0.0f;
		move += offset;

		Debug.DrawRay (move, Vector3.up * 2, Color.red, 1f);

		if (collisionCount > normalCollisionCount)
			move = move.normalized * collisionMoveFactor;
		else if (move.sqrMagnitude > sqrMoveLimit)
			move = move.normalized * moveLimit;
				
		Vector3 finalPos = myRigidbody.position + move;

		Debug.DrawRay (finalPos, Vector3.up * 2, Color.yellow, 1f);

		myRigidbody.MovePosition (finalPos); 	
	}

		
	public override void FixEvents ()
	{
		base.FixEvents ();
		//myRigidbody.isKinematic = false;
	}


	public override void RemoveEvents ()
	{
		base.RemoveEvents ();
		//myRigidbody.isKinematic = true;
	}



	public void ResetSettings ()
	{
		collisionCount = 0;
	}

	public void FreezePhysics ()
	{
		Debug.Log (string.Format ("DragablePhysics - FreezePhysics ()"));

		myRigidbody.useGravity = false;
		myRigidbody.isKinematic = true;
	}

}
