﻿using UnityEngine;
using AudioSystem;
using System.Collections;

/// <summary>
/// Particle controller is a utilitie script that can be added to particles to become part of the game architecture.
/// The class will deactivate all the particles(if selected show) and make the available to the "EnableObjective" And "DisableObjective" messages
/// </summary>
public class ParticleController : TaskHandler
{
	/// <summary>
	/// The particle list container.
	/// </summary>
	ParticleSystem[] particles;

	/// <summary>
	/// If all particles should be deactivated
	/// </summary>
	public bool stopAllAtStartFires = false;

	private bool particlesAreAPlaying = true;

	/// <summary>
	/// if upon deactivation the particle should do a rewind effect
	/// </summary>
	public bool rewindBackInTime = false;
		
	private float timecount = 0;

	void Awake ()
	{	
		//Get All the particle of the Game object and its children
		particles = GetComponentsInChildren <ParticleSystem> ();
		myTransform = transform;
		if (stopAllAtStartFires)
			DisableObjective ();
	}

	/// <summary>
	/// Enables the objective and therfore the particles.
	/// </summary>
	public void EnableObjective ()
	{
		particlesAreAPlaying = true;
		foreach (ParticleSystem partItem in particles)
			partItem.Play ();
	}

	/// <summary>
	/// Disables the Particles.
	/// </summary>
	public void DisableObjective ()
	{
		particlesAreAPlaying = false;
				
		particles = GetComponentsInChildren <ParticleSystem> ();

		foreach (ParticleSystem partItem in particles)
			partItem.Stop ();

		//if the rewind effect is set
		if (rewindBackInTime) {
			//iterate all the particles and the most largest play time
			foreach (ParticleSystem partItem in particles)
				if (timecount < partItem.time)
					timecount = partItem.time;
			//Start the rewind effect
			StartCoroutine (RewindParticleSystem ());
		}
	}

	IEnumerator RewindParticleSystem ()
	{
		//While the time is above 0
		while (timecount > 0.01) {
			// reduce the play time
			timecount = Mathf.Lerp (timecount, 0, 0.2f); 
			//Simulate the particle
			foreach (ParticleSystem partItem in particles)
				partItem.Simulate (timecount, false, true);
						
			yield return new WaitForFixedUpdate ();
			yield return 1;
		}
	}

	/// <summary>
	/// Toggle the particle mode.
	/// </summary>
	public void Toggle ()
	{
		if (particlesAreAPlaying)
			DisableObjective ();
		else
			EnableObjective ();
	}
		
	//if particles are part of a task
	public override void EnableTask ()
	{
		EnableObjective ();
		StartCoroutine (CompleteTheTask ());
	}

	//Send a task complete message
	IEnumerator CompleteTheTask ()
	{
		yield return StartCoroutine (Auto.Wait (HRUtilities.at));
		SendMessageUpwards (HRUtilities.TaskCompleted, this.gameObject, SendMessageOptions.DontRequireReceiver);
	}

		
	private Transform _myTransform;

	public Transform myTransform {
		get {
			if (_myTransform == null)
				_myTransform = GetComponent<Transform> ();
			return _myTransform;
		}
		set {
			_myTransform = value;
		}
	}
}